/*    WIPort 2 RESources implementation
 *
 *    Copyright (c) 1994--1998 Alexandru Dan Corlan
 * 
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU Library General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    In short, you can use and modify this program. You can
 *    distribute the program with or without your modifications if you want,
 *    but you must not require licence fees and must provide the
 *    full source code with the distribution. If you make an application program
 *    which just calls functions from this library you may distribute that
 *    application program without source code, but also as an linkable object file, 
 *    and must distribute also this library with source code.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* key definitions  */
#define C_A        0x0001
#define C_B        0x0002
#define C_C        0x0003
#define C_D        0x0004
#define C_E        0x0005
#define C_F        0x0006
#define C_G        0x0007
#define C_H        0x0008
#define C_I        0x0009
#define C_J        0x000a
#define C_K        0x000b
#define C_L        0x000c
#define C_M        0x000d
#define C_N        0x000e
#define C_O        0x000f
#define C_P        0x000f
#define C_Q        0x0010
#define C_R        0x0011
#define C_S        0x0012
#define C_T        0x0013
#define C_U        0x0014
#define C_V        0x0015
#define C_X        0x0016
#define C_Y        0x0017
#define C_Z        0x0018
#define ESC        0x001b
#define DEL        0x0004

#ifdef DEFINE_KEY_NAMES
char * special_keys[] = {
  "BS", "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", 
  "F11", "F12", "UP", "DOWN", "LEFT", "RIGHT", "TAB","INS", "HOME",
  "END", "PGUP", "PGDOWN"
};
#endif

#define F1       0x0101
#define F2       0x0102
#define F3       0x0103
#define F4       0x0104
#define F5       0x0105
#define F6       0x0106
#define F7       0x0107
#define F8       0x0108
#define F9       0x0109
#define F10      0x010a
#define F11      0x010b
#define F12      0x010c
#define UP       0x0010
#define DOWN     0x000e
#define LEFT     0x0002
#define RIGHT    0x0005
#define TAB      0x0111
#define INS      0x0112
#define HOME     0x0001
#define END      0x0114
#define PGUP     0x0115
#define PGDOWN   0x0116
#define KPD_SLASH  0x0117
#define KPD_STAR   0x0118
#define KPD_MINUS  0x0119
#define KPD_PLUS   0x011a
#define KPD_ENTER  0x011b
#define KPD_0      0x011c
#define KPD_1      0x011d
#define KPD_2      0x011e
#define KPD_3      0x011f
#define KPD_4      0x0120
#define KPD_5      0x0121
#define KPD_6      0x0122
#define KPD_7      0x0123
#define KPD_8      0x0124
#define KPD_9      0x0125
#define BS         0x007f

#define WIP_SHIFT  0x0200 
#define S_F1       0x0201
#define S_F2       0x0202
#define S_F3       0x0203
#define S_F4       0x0204
#define S_F5       0x0205
#define S_F6       0x0206
#define S_F7       0x0207
#define S_F8       0x0208
#define S_F9       0x0209
#define S_F10      0x020a
#define S_F11      0x020b
#define S_F12      0x020c
#define S_UP       0x020d
#define S_DOWN     0x020e
#define S_LEFT     0x020f
#define S_RIGHT    0x0210
#define S_TAB      0x0211
#define S_INS      0x0212
#define S_HOME     0x0213
#define S_END      0x0214
#define S_PGUP     0x0215
#define S_PGDOWN   0x0216
#define S_K_SLASH  0x0217
#define S_K_STAR   0x0218
#define S_K_MINUS  0x0219
#define S_K_PLUS   0x021a
#define S_K_ENTER  0x021b
#define S_K_0      0x021c
#define S_K_1      0x021d
#define S_K_2      0x021e
#define S_K_3      0x021f
#define S_K_4      0x0220
#define S_K_5      0x0221
#define S_K_6      0x0222
#define S_K_7      0x0223
#define S_K_8      0x0224
#define S_K_9      0x0225
#define S_BS       0x0230

#define WIP_CTRL   0x4000
#define C_F1       0x0401
#define C_F2       0x0402
#define C_F3       0x0403
#define C_F4       0x0404
#define C_F5       0x0405
#define C_F6       0x0406
#define C_F7       0x0407
#define C_F8       0x0408
#define C_F9       0x0409
#define C_F10      0x040a
#define C_F11      0x040b
#define C_F12      0x040c
#define C_UP       0x040d
#define C_DOWN     0x040e
#define C_LEFT     0x040f
#define C_RIGHT    0x0410
#define C_TAB      0x0411
#define C_INS      0x0412
#define C_HOME     0x0413
#define C_END      0x0414
#define C_PGUP     0x0415
#define C_PGDOWN   0x0416
#define C_K_SLASH  0x0417
#define C_K_STAR   0x0418
#define C_K_MINUS  0x0419
#define C_K_PLUS   0x041a
#define C_K_ENTER  0x041b
#define C_K_0      0x041c
#define C_K_1      0x041d
#define C_K_2      0x041e
#define C_K_3      0x041f
#define C_K_4      0x0420
#define C_K_5      0x0421
#define C_K_6      0x0422
#define C_K_7      0x0423
#define C_K_8      0x0424
#define C_K_9      0x0425
#define C_BS       0x0430

#define WIP_ALT    0x0800
#define A_F1       0x0801
#define A_F2       0x0802
#define A_F3       0x0803
#define A_F4       0x0804
#define A_F5       0x0805
#define A_F6       0x0806
#define A_F7       0x0807
#define A_F8       0x0808
#define A_F9       0x0809
#define A_F10      0x080a
#define A_F11      0x080b
#define A_F12      0x080c
#define A_UP       0x080d
#define A_DOWN     0x080e
#define A_LEFT     0x080f
#define A_RIGHT    0x0810
#define A_TAB      0x0811
#define A_INS      0x0812
#define A_HOME     0x0813
#define A_END      0x0814
#define A_PGUP     0x0815
#define A_PGDOWN   0x0816
#define A_K_SLASH  0x0817
#define A_K_STAR   0x0818
#define A_K_MINUS  0x0819
#define A_K_PLUS   0x081a
#define A_K_ENTER  0x081b
#define A_K_0      0x081c
#define A_K_1      0x081d
#define A_K_2      0x081e
#define A_K_3      0x081f
#define A_K_4      0x0820
#define A_K_5      0x0821
#define A_K_6      0x0822
#define A_K_7      0x0823
#define A_K_8      0x0824
#define A_K_9      0x0825
#define A_BS       0x0830
